// Klax.cpp: Implementierung der Klasse CKlax.
//
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include "Klax.h"

//////////////////////////////////////////////////////////////////////
// Konstanten
//////////////////////////////////////////////////////////////////////
#define STEINHOEHE 15
#define STEINBREITE 50
#define STEINPUNKTE 5

#define SPIEL_NEU 0
#define SPIEL_LAEUFT 1
#define SPIEL_FEHLER 2
#define SPIEL_ENDE 3
#define SPIEL_PAUSE 4
#define SPIEL_LEVELWECHSEL 5

const COLORREF CKlax::farbtabelle[2][8] = {
	RGB(0,0,0),			// Scharz
	RGB(255,255,255),	// Wei
	RGB(255,0,0),		// Rot
	RGB(0,255,0),		// Grn
	RGB(0,0,255),		// Blau
	RGB(255,255,0),		// Gelb
	RGB(255,0,255),		// Magenta
	RGB(0,255,255),		// Cyan
	RGB(0,0,0),			// Scharz
	RGB(180,180,180),	// Grau
	RGB(180,0,0),		// DunkelRot
	RGB(0,180,0),		// DunkelGrn
	RGB(0,0,180),		// DunkelBlau
	RGB(180,180,0),		// DunkelGelb
	RGB(180,0,180),		// DunkelMagenta
	RGB(0,180,180)		// DunkelCyan
	};

const int CKlax::punkte[8] = { 
	STEINPUNKTE,	// Stein aufnehmen	(so def. wegen Spielerstructur)
	STEINPUNKTE*10,	// 3 Steine Senkrecht
	STEINPUNKTE*20,	// 3 Steine Wagerecht
	STEINPUNKTE*40,	// 3 Steine Diagonal / 4 Steine Senkrecht
	STEINPUNKTE*80,	// 4 Steine Wagerecht
	STEINPUNKTE*160,// 4 Steine Diagonal
	STEINPUNKTE*320,// 5 Steine Wagerecht
	STEINPUNKTE*640 // 5 Steine Diagonal
	};




//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

CKlax::CKlax()
{

}

CKlax::~CKlax()
{

}

//////////////////////////////////////////////////////////////////////
// Funktionen -> Public
//////////////////////////////////////////////////////////////////////

void CKlax::gameloop()
{
	if (status==SPIEL_LAEUFT)
	{
		static int frame;
		frame++;

		if ((player.fehler>0)&&(player.leben>0))
		{
			if (frame==(10-player.level))
			{
				durchlauf[int(rand()%4)][0] = (1+(rand()%optionen.farben));
				frame = 0;
			}
			durchlauf_weiter();
		}
		else
		{
			if(player.leben>0)
			{
				status = SPIEL_FEHLER;
			}
			else
			{
				status = SPIEL_ENDE;
				spiel_gameover();
			}
		}

		if (player.levelziel[player.level]<1)
		{	
			status = SPIEL_LEVELWECHSEL;
		}
	}
	
	if (status==SPIEL_LEVELWECHSEL)
	{
		// Durchlauf und Halter lschen
		ZeroMemory(&durchlauf,sizeof(durchlauf));
		ZeroMemory(&player.halter,sizeof(player.halter));

		// Punkte fr freie felder im Auffang
		int i,j;
		for (i=0;i<5;i++)
		{	for (j=0;j<5;j++)
			{
				if (auffang[i][j]==0)
				{
					player.punkte+=50;
				}
				else
				{
					auffang[i][j]=0;
				}
			}
		}
		player.level++;
		status=SPIEL_LAEUFT;
		if (player.level==8) status=SPIEL_ENDE;
	}

}

void CKlax::spiel_init(int geschw, int fehler, int leben, int farben, bool lifo)
{
	// Spielparameter setzten
	optionen.farben = farben;
	optionen.fehler = fehler;
	optionen.geschw = geschw;
	optionen.leben = leben;
	optionen.lifo = lifo;

	// Spieler initialisieren
	spieler_init();

	// Felder mit Nullen fllen
	ZeroMemory(&durchlauf,sizeof(durchlauf));
	ZeroMemory(&auffang,sizeof(auffang));

	// Zufallszahlen Generator init
	srand(GetTickCount());

	// Zeichenmittel erzeugen
	int i;
	for (i=0;i<7;i++)
	{	// Stiffte
		stifft[i] = CreatePen(PS_SOLID,2,farbtabelle[1][i+1]);
		// Pinsel
		pinsel[i] = CreateSolidBrush(farbtabelle[0][i+1]);
	}

	textFont = CreateFont(15,8,0,0,FW_NORMAL,0,0,0,
		DEFAULT_CHARSET,OUT_CHARACTER_PRECIS,CLIP_CHARACTER_PRECIS,DEFAULT_QUALITY,
		DEFAULT_PITCH|FF_DONTCARE, "Comic Sans MS");

	status = SPIEL_NEU;
	
}


//////////////////////////////////////////////////////////////////////
// Funktionen -> Privat -> Spielfunktionen
//////////////////////////////////////////////////////////////////////

void CKlax::spiel_start()
{
	status = SPIEL_LAEUFT;
}

void CKlax::spiel_gameover()
{

}

void CKlax::spiel_weiter()
{
	player.leben--;
	player.fehler = optionen.fehler;

	ZeroMemory(&durchlauf,sizeof(durchlauf));
	ZeroMemory(&auffang,sizeof(auffang));
	ZeroMemory(&player.halter,sizeof(player.halter));

	status = SPIEL_LAEUFT;
}

void CKlax::spiel_pause()
{
	if (status==SPIEL_PAUSE)
	{
		status = SPIEL_LAEUFT;
	}
	else 
	{
		status = SPIEL_PAUSE;
	}
}

void CKlax::spiel_zeichnen(HDC &hdc)
{
	hintergrund_zeichnen(hdc);
	
	if (status==SPIEL_LAEUFT)
	{
		durchlauf_zeichnen(hdc);
		auffang_zeichnen(hdc);
	}
	spieler_zeichnen(hdc);
	texte_zeichnen(hdc);
}


//////////////////////////////////////////////////////////////////////
// Funktionen -> Privat -> Spieler
//////////////////////////////////////////////////////////////////////

void CKlax::spieler_init()
{
	// Spieloptionen auf Spieler bertragen
	player.fehler = optionen.fehler;
	player.leben = optionen.leben;
	player.pos = 2;
	player.punkte = 0;
	player.level = 1;
	ZeroMemory(&player.halter,sizeof(player.halter));
	player.level_init();
}

void CKlax::spieler_action()
{
	switch (status)
	{
		case SPIEL_NEU: spiel_start();break;
		case SPIEL_LAEUFT:
			{
				if (player.halter[0]>0)
				{
					auffang_neu(player.pos, player.stein_abgeben(optionen.lifo));
					auffang_testen();
				}
			} break;
		case SPIEL_FEHLER: spiel_weiter(); break;
	}
}

void CKlax::spieler_bewegen(bool richtung)
{
	// richtung = true dann rechts sonst links
	if (richtung)
	{
		if (player.pos<4)player.pos++;
	}
	else
	{
		if (player.pos>0)player.pos--;
	}
}

byte CKlax::Spieler::stein_abgeben(bool lifo)
{
	byte tmp;

	if (lifo)
	{
		tmp = halter[halter[0]];	// Wert zum abgeben zwischenspeichern
		halter[halter[0]] = 0;		// Ursprungswert lschen
	}
	else
	{
		tmp = halter[1];	// Wert zum abgeben zwischenspeichern
		halter[1]=halter[2];// anderen werte durchrcken
		halter[2]=halter[3];
		halter[3]=0;
	}
	halter[0]--;		// gre neu einstellen
	return tmp;			// Wert ausgeben
}

void CKlax::Spieler::stein_aufnehmen(byte farbe)
{
	int i;
	for (i=1;i<4;i++)
	{
		if (halter[i]==0) {halter[i]=farbe; halter[0]=i;break;}
	}
	punkte+=STEINPUNKTE;
	leveltest(0);
}

void CKlax::Spieler::leveltest(byte quelle)
{
	switch(level)
	{
		case 1:
			if ((quelle==1)||(quelle==2)) levelziel[1]--;
			break;
		case 2:
			if ((quelle==2)||(quelle==4)||(quelle==6)) levelziel[2]--;
			break;
		case 3:
			if ((quelle==3)||(quelle==5)||(quelle==7)) levelziel[3]--;
			break;
		case 4:
			if (quelle==0) levelziel[4]--;
			break;
		case 5:
			{
				static long startpunkte;
				if (startpunkte<1) startpunkte = punkte;

				long dif = punkte - startpunkte;
				if ((levelziel-dif)<0)
				{
					levelziel[5]=0;
					startpunkte=0;
				}
				else
				{
					levelziel[5]=levelziel[5]-dif;
					startpunkte+=dif;
				}
			}break;
		case 6:
			if (quelle==6) levelziel[6]--;
		case 7:
			if (quelle==7) levelziel[7]--;
			break;
	}
}

void CKlax::Spieler::level_init()
{
	levelziel[0]=0;		// noch keine bedeutung
	levelziel[1]=5;		// Steine Senkrecht
	levelziel[2]=5;		// Steine Wagerecht
	levelziel[3]=5;		// Steine Diagonal
	levelziel[4]=40;	// Steine Auffangen
	levelziel[5]=1000;	// Punkte erreichen
	levelziel[6]=5;		// 5 Steine Wagerecht
	levelziel[7]=3;		// 5 Steine Diagonal
}


//////////////////////////////////////////////////////////////////////
// Funktionen -> Privat -> Felder
//////////////////////////////////////////////////////////////////////

void CKlax::durchlauf_weiter()
{
	byte auswurf=5;
	int i,j;

	// alles weiterrcken
	for (i=0;i<5;i++)
	{
		for (j=11;j>0;j--)
		{
			durchlauf[i][j] = durchlauf[i][j-1];
		}
		durchlauf[i][0]=0;
		if (durchlauf[i][11] > 0) auswurf=i;
	}

	// auf Auswurf achten
	if (auswurf<5)
	{
		if (player.pos==auswurf) 
		{
			player.stein_aufnehmen(durchlauf[auswurf][11]);
		}
		else
		{
			player.fehler--;
			if (player.fehler==0) player.leben--;
		}
	}

}

void CKlax::auffang_neu(byte pos, byte farbe)
{
	if (auffang[pos][0]==0)
	{
		int i;
		for (i=4;i>=0;i--)
		{
			if (auffang[pos][i]==0) {auffang[pos][i]=farbe;break;}
		}
		auffang_testen();
	}
	else
	{
		player.fehler--;
		if (player.fehler==0) player.leben--;
	}
}

void CKlax::auffang_testen(byte multi)
{
	int i,j;

	// hier wird getestet ob Reihen entstanden sind

	// Steine Diagonal
	for (i=0;i<5;i++)
	{	for (j=0;j<3;j++)
		{	if(auffang[i][j]>0)
			{	if(i<=2)	// Linke Seite und Mitte
				{	if((auffang[i][j]==auffang[i+1][j+1])&&(auffang[i][j]==auffang[i+2][j+2]))
					{	if(auffang[i][j]==auffang[i+3][j+3])
						{	if(auffang[i][j]==auffang[i+4][j+4])
							{
								player.punkte+=(punkte[7]*multi);
								auffang[i][j]=auffang[i+1][j+1]=auffang[i+2][j+2]=auffang[i+3][j+3]=auffang[i+4][j+4]=0;
								auffang_nachruecken();auffang_testen(multi+1);
								player.leveltest(7);
								break;
							}
							player.punkte+=(punkte[5]*multi);
							auffang[i][j]=auffang[i+1][j+1]=auffang[i+2][j+2]=auffang[i+3][j+3]=0;
							auffang_nachruecken();auffang_testen(multi+1);
							player.leveltest(5);
							break;
						}
						player.punkte+=(punkte[3]*multi);
						auffang[i][j]=auffang[i+1][j+1]=auffang[i+2][j+2]=0;
						auffang_nachruecken();auffang_testen(multi+1);
						player.leveltest(3);
						break;
					}
				}

				if(i>=2)	// Rechte Seite und Mitte
				{	if((auffang[i][j]==auffang[i-1][j+1])&&(auffang[i][j]==auffang[i-2][j+2]))
					{	if(auffang[i][j]==auffang[i-3][j+3])
						{	if(auffang[i][j]==auffang[i-4][j+4])
							{
								player.punkte+=(punkte[7]*multi);
								auffang[i][j]=auffang[i-1][j+1]=auffang[i-2][j+2]=auffang[i-3][j+3]=auffang[i-4][j+4]=0;
								auffang_nachruecken();auffang_testen(multi+1);
								player.leveltest(7);
								break;
							}
							player.punkte+=(punkte[5]*multi);
							auffang[i][j]=auffang[i-1][j+1]=auffang[i-2][j+2]=auffang[i-3][j+3]=0;
							auffang_nachruecken();auffang_testen(multi+1);
							player.leveltest(5);
							break;
						}
						player.punkte+=(punkte[3]*multi);
						auffang[i][j]=auffang[i-1][j+1]=auffang[i-2][j+2]=0;
						auffang_nachruecken();auffang_testen(multi+1);
						player.leveltest(3);
						break;
					}
				}
			}
		}
	}


	// Steine Wagerecht
	for (j=0;j<5;j++)
	{	for (i=0;i<3;i++)
		{	if(auffang[i][j]>0)
			{
				if(auffang[i][j]==auffang[i+1][j])
				{	if(auffang[i][j]==auffang[i+2][j])
					{	if(i<2) 
						{	if(auffang[i][j]==auffang[i+3][j])
							{	if(i==0)
								{	if(auffang[i][j]==auffang[i+4][j])
									{
										player.punkte+=(punkte[6]*multi);
										auffang[i][j]=auffang[i+1][j]=auffang[i+2][j]=auffang[i+3][j]=auffang[i+4][j]=0;
										auffang_nachruecken();auffang_testen(multi+1);
										player.leveltest(6);
										break;
									}
								}
								player.punkte+=(punkte[4]*multi);
								auffang[i][j]=auffang[i+1][j]=auffang[i+2][j]=auffang[i+3][j]=0;
								auffang_nachruecken();auffang_testen(multi+1);
								player.leveltest(4);
								break;
							}
						}
						player.punkte+=(punkte[2]*multi);
						auffang[i][j]=auffang[i+1][j]=auffang[i+2][j]=0;
						auffang_nachruecken();auffang_testen(multi+1);
						player.leveltest(2);
					}
				}
			}
		}
	}


	// Steine Senkrecht
	for (i=0;i<5;i++)
	{	for (j=0;j<3;j++)
		{	if(auffang[i][j]>0)
			{
				if(auffang[i][j]==auffang[i][j+1])
				{	if(auffang[i][j]==auffang[i][j+2])
					{	if(j<2) 
						{	if(auffang[i][j]==auffang[i][j+3])
							{
								player.punkte+=(punkte[3]*multi);
								auffang[i][j]=auffang[i][j+1]=auffang[i][j+2]=auffang[i][j+3]=0;
								player.leveltest(3);
								break;
							}
						}
						player.punkte+=(punkte[1]*multi);
						auffang[i][j]=auffang[i][j+1]=auffang[i][j+2]=0;
						player.leveltest(1);
					}
				}
			}
		}
	}
}

void CKlax::auffang_nachruecken()
{
	int i,j;
	for (i=0;i<5;i++)
	{	for (j=4;j>=1;j--)
		{
			if (auffang[i][j]==0)
			{
				auffang[i][j]=auffang[i][j-1];
				auffang[i][j-1]=0;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////
// Funktionen -> Privat -> zeichnen
//////////////////////////////////////////////////////////////////////

void CKlax::stein_zeichnen(HDC &hdc, int posx, int posy, byte farbe)
{         
	SelectObject(hdc,stifft[farbe-1]);
	SelectObject(hdc,pinsel[farbe-1]);
	RoundRect(hdc,posx,posy,posx+STEINBREITE,posy+STEINHOEHE,30,15);
}

void CKlax::spieler_zeichnen(HDC &hdc)
{   
	if (player.halter[0] >0)
	{
		byte i;
		for (i=1;i<=player.halter[0];i++)
		{
			stein_zeichnen(hdc,50+(60*player.pos),260+((player.halter[0]-i)*15),player.halter[i]);
		}
	}

	SelectObject(hdc,stifft[0]);
	SelectObject(hdc,pinsel[0]);
	RoundRect(hdc,40+(60*player.pos),265+(player.halter[0]*15),
				  40+(60*player.pos)+STEINBREITE+20,265+(player.halter[0]*15)+STEINHOEHE-5,30,15);
}

void CKlax::hintergrund_zeichnen(HDC &hdc)
{   
	// Trennlienie Anzeige - Spielfeld
	SelectObject(hdc,stifft[1]);
	MoveToEx(hdc,5,50,NULL);
	LineTo(hdc,390,50);

	// durchlauf - auswurf
	MoveToEx(hdc,5,220,NULL);
	LineTo(hdc,390,220);

	// Auffang
	SelectObject(hdc, stifft[1]);
	SelectObject(hdc, pinsel[1]);
	Rectangle(hdc,30,340,40,455);	// Links
	Rectangle(hdc,350,340,360,455);	// Rechts
	Rectangle(hdc,30,445,360,455);	// Unten
}

void CKlax::durchlauf_zeichnen(HDC &hdc)
{
	int i,j;
	for (i=0;i<5;i++)
	{	for (j=0;j<11;j++)
		{	
			if (durchlauf[i][j]>0) stein_zeichnen(hdc,55+(60*i),55+(15*j),durchlauf[i][j]);
		}
	}
}

void CKlax::auffang_zeichnen(HDC &hdc)
{
	int i,j;
	for (i=0;i<5;i++)
	{	for (j=0;j<5;j++)
		{	
			if (auffang[i][j]>0) stein_zeichnen(hdc,50+(60*i),350+(18*j),auffang[i][j]);
		}
	}
}

double CKlax::potenz(int zahl, int potenz)
{
	int i;
	double tmp=1;
	if (potenz==0)return tmp;
	for (i=1;i<=potenz;i++)
	{
		tmp=zahl*tmp;
	}
	return tmp;
}

void CKlax::texte_zeichnen(HDC &hdc)
{
	int i;

	SetBkMode(hdc,TRANSPARENT);
	SetTextColor(hdc, farbtabelle[0][2]);
	SelectObject(hdc,textFont);
	
	// Punkte umrechnen in ASCII
	char Score[] = "00000000";
	
	char tmp[8]={0};
	double teiler;
	double tmpScore = player.punkte;

	for (i=0;i<7;i++)
	{
		teiler = potenz(10,7-i);
		tmp[i] = int(tmpScore/teiler);
		if(tmp[i]>0)
		{
			tmpScore = tmpScore-tmp[i]*teiler;
			Score[i] = tmp[i]+48;
		}
	}
	Score[7] = int(tmpScore)+48;
	
	// Leben
	char Leben[] = "Leben: X";
	Leben[7] = player.leben+48;

	// Level
	char Level[] = "Level: X";
	Level[7] = player.level+48;

	// Levelziel
	char Ziel[] = "Ziel: 00000 x";
	if (player.level==5)
	{
		char tmp[5]={0};
		double teiler;
		double puffer = player.levelziel[5];

		for (i=0;i<5;i++)
		{
			teiler = potenz(10,4-i);
			tmp[i] = int(puffer/teiler);
			if(tmp[i]>0)
			{
				puffer = puffer-tmp[i]*teiler;
				Ziel[i+5] = tmp[i]+48;
			}	
		}
	}
	else
	{
		int puffer = player.levelziel[player.level];
		Ziel[6]=int(puffer/10)+48;
		Ziel[7]=puffer-(int(puffer/10)*10)+48;
		Ziel[8]=' '; Ziel[9]=' '; Ziel[10]=' ';
	}


	char Zieltext[19]={" "};
	char tmp1[] = "Steine Senkrecht";
	char tmp2[] = "Steine Wagerecht";
	char tmp3[] = "Steine Diagonal";
	char tmp4[] = "Steine auffangen";
	char tmp5[] = "Punkte ereichen";
	char tmp6[] = "5 Steine Wagerecht";
	char tmp7[] = "5 Steine Diagonal";

	switch (player.level)
	{
		case 1:CopyMemory(&Zieltext,&tmp1, sizeof(tmp1));break;
		case 2:CopyMemory(&Zieltext,&tmp2, sizeof(tmp2));break;
		case 3:CopyMemory(&Zieltext,&tmp3, sizeof(tmp3));break;
		case 4:CopyMemory(&Zieltext,&tmp4, sizeof(tmp4));break;
		case 5:CopyMemory(&Zieltext,&tmp5, sizeof(tmp5));break;
		case 6:CopyMemory(&Zieltext,&tmp6, sizeof(tmp6));break;
		case 7:CopyMemory(&Zieltext,&tmp7, sizeof(tmp7));break;
	}

	// Mgliche Fehler
	char Fehler[] = "";
	
	for (i=1;i<=(10-player.fehler);i++)
	{
		 Fehler[i-1]= ' ';
	}

	// Texte ausgeben
	SetTextAlign(hdc, TA_LEFT);
	TextOut(hdc, 5, 10, Leben, lstrlen(Leben));
	TextOut(hdc, 5, 30, Level, lstrlen(Level));
	TextOut(hdc, 85, 10, Ziel, lstrlen(Ziel));
	TextOut(hdc, 85, 30, Zieltext, lstrlen(Zieltext));

	SetTextAlign(hdc, TA_RIGHT);
	TextOut(hdc, 390, 10, Score, lstrlen(Score));
	TextOut(hdc, 390, 30, Fehler, lstrlen(Fehler));
	
	SetTextAlign(hdc, TA_CENTER);
	
	// Statustexte
	char Spielstart[] = "Start mit Leertaste.";
	
	switch (status) 
	{
		case SPIEL_FEHLER:
		{
			char zuvielFehler1[] = "Sie haben zu viele Steine Fallen gelassen!";
			char zuvielFehler2[] = "Noch einmal Probieren?";
			TextOut(hdc, 200, 80, zuvielFehler1, lstrlen(zuvielFehler1));
			TextOut(hdc, 200, 100, zuvielFehler2, lstrlen(zuvielFehler2));
			TextOut(hdc, 200, 120, Spielstart, lstrlen(Spielstart));
		} break;
		case SPIEL_NEU:
			TextOut(hdc, 200, 120, Spielstart, lstrlen(Spielstart));
			break;
		case SPIEL_PAUSE:
		{
			char pause[] = "P A U S E";
			static bool blink;
			blink=!blink;
			if (blink) TextOut(hdc, 200, 120, pause, lstrlen(pause));
		} break;
		case SPIEL_ENDE:
			char ende[]= "S P I E L   E N D E";
			TextOut(hdc, 200, 120, ende, lstrlen(ende));
			break;
	}
}